(******************************************************************************)
(* 1. Modules z'et foncteurs **************************************************)
(******************************************************************************)

(* 1.1. Compteur **************************************************************)

(* Question 1. *)

let c = ref 0;;
let clear () = c := 0;;
let incr  () = c := !c + 1;;
let value () = !c;;

clear ();;
incr ();;
value ();;
c := 1664;;
value ();;


(* Question 2. *)

let (clear, incr, value) =
  let c = ref 0 in
    ((fun () -> c := 0),
     (fun () -> c := !c + 1),
     (fun () -> !c));;
     
clear ();;
incr ();;
value ();;
c := 1664;;
value ();;


(* Question 3. *)

module Counter =
  (struct
     let c = ref 0
     let clear () = c := 0
     let incr  () = c := !c + 1
     let value () = !c
   end : sig
     val clear : unit -> unit
     val incr  : unit -> unit
     val value : unit -> int
   end);;

Counter.clear ();;
Counter.incr ();;
Counter.value ();;
Counter.c := 1664;;
Counter.value ();;



(* 1.2. Structures algbriques : les anneaux **********************************)

module type RING =
sig
  type t
  val zero  : t                (* 0             *)
  val one   : t                (* 1             *)
  val add   : t -> t -> t      (* x y ->  x + y *)
  val opp   : t -> t           (* x   -> -x     *)
  val mult  : t -> t -> t      (* x y ->  x * y *)
  val print : t -> unit
end;;

(* Question 4. *)

module ZRing : (RING with type t = int) =
struct
  type t = int
  let zero = 0
  let one  = 1
  let add  x y =  x + y
  let opp  x   = -x
  let mult x y =  x * y
  let print x  = print_int x
end;;

module R = ZRing;;
R.print (R.mult (R.add 4 4) (R.opp (R.mult (R.add R.one 51) 4)));;


(* Question 5. *)

module BRing : (RING with type t = bool) =
struct
  type t = bool
  let zero = false
  let one  = true
  let add  x y =  x <> y
  let opp  x   =  x
  let mult x y =  x && y
  let print x  = print_string (if x then "I" else "O")
end;;

module R = BRing;;
R.print (R.mult true (R.add false true));;


(* Question 6. *)

module CRing : (RING with type t = float * float) =
struct
  type t = float * float
  let zero = (0., 0.)
  let one  = (1., 0.)
  let add  (rx,ix) (ry,iy) = (rx+.ry, ix+.iy)
  let opp  (rx,ix) = (-.rx,-.ix)
  let mult (rx,ix) (ry,iy) = (rx*.ry -. ix*.iy, rx*.iy +. ix*.ry)
  let print (rx,ix) =
    print_float rx; print_string " + ";
    print_float ix; print_string "i"
end;;

module R = CRing;;
R.print (R.mult (R.add (2.,0.) (14.,0.)) (R.opp (R.add (-1.,3.) R.one)));;


(* Question 7. *)

module SeqRing (A : RING) : (RING with type t = int -> A.t) =
struct
  type t = int -> A.t
  let zero = fun n -> A.zero
  let one  = fun n -> A.one
  let add  u1 u2 = fun n -> A.add  (u1 n) (u2 n)
  let opp  u     = fun n -> A.opp  (u  n)
  let mult u1 u2 = fun n -> A.mult (u1 n) (u2 n)
  let print u =
    print_string "[ ";
    for i = 0 to 9 do
      if i > 0 then print_string " ; ";
      A.print (u i)
    done;
    print_string " ]"
end;;

module R = SeqRing(ZRing);;
R.print (R.mult (fun n -> 4*n) (fun n -> 2));;


(* Question 8. *)

module ProdRing (A : RING) (B : RING) : (RING with type t = A.t * B.t) =
struct
  type t = A.t * B.t
  let zero = (A.zero, B.zero)
  let one = (A.one, B.one)
  let add  (a1,b1) (a2,b2) = (A.add  a1 a2, B.add  b1 b2)
  let opp (a,b) = (A.opp a, B.opp b)
  let mult (a1,b1) (a2,b2) = (A.mult a1 a2, B.mult b1 b2)
  let print (a,b) =
    print_string "( "; A.print a;
    print_string " , "; B.print b;
    print_string " )"
end;;

module R = ProdRing(ZRing)(CRing);;
R.print (R.mult (16, (6.,4.)) (51, (4.,2.)));;


(* Question 9. *)

module PolyRing (A : RING) : (RING with type t = A.t list) =
struct
  type t = A.t list
  let zero = [A.zero]
  let one = [A.one]
  let rec add p1 p2 =
    match (p1,p2) with
      | ([],[]) -> []
      | (_,[]) -> p1
      | ([],_) -> p2
      | (x1::q1,x2::q2) -> (A.add x1 x2)::(add q1 q2)
  let rec opp p =
    match p with
      | [] -> []
      | x::q -> (A.opp x)::(opp q)
  let rec mult_k p k =
    match p with
      | [] -> []
      | x::q -> (A.mult x k)::(mult_k q k)
  let rec mult p1 p2 =
    match p2 with
      | [] -> []
      | x2::q2 -> (add (mult_k p1 x2) (A.zero::(mult p1 q2)))
  let print p =
    let rec print_aux p n = match p with
      | [] -> ()
      | x::q ->
	  if n > 0 then
	    print_string " + ";
	  A.print x;
	  if n > 0 then print_string " * x";
	  if n > 1 then (print_string "^"; print_int n);
	  print_aux q (n+1)
    in print_aux p 0
end;;

module R = PolyRing(ZRing);;
R.print (R.mult [5;1] [16;-6;4]);;
